{% from 'macros/common_macros.lua.j2' import boolean, quotes, qtype_table %}


{% macro cache_clear(params) -%}
cache.clear(
{{- quotes(params.name) if params.name else 'nil' -}},
{{- boolean(params.exact_name) -}},
{{- qtype_table(params.rr_type) if params.rr_type else 'nil' -}},
{{- params.chunk_size if not params.exact_name else 'nil' -}}
)
{%- endmacro %}
