{% from 'macros/common_macros.lua.j2' import boolean %}

{% if cfg.defer.enable and disable_defer is not defined -%}
assert(C.defer_init(
	'{{ cfg.rundir }}/defer',
	{{ cfg.defer.log_period.millis() }},
	{{ cfg.workers }}) == 0)
{% else %}
assert(C.defer_init(nil, 0, 0) == 0)
{%- endif %}
