{%- macro get_proto_set(protocols) -%}
0
{%- for p in protocols or [] -%}
 + 2^C.KR_PROTO_{{ p.upper() }}
{%- endfor -%}
{%- endmacro -%}

{% macro view_flags(options) -%}
{% if not options.minimize -%}
"NO_MINIMIZE",
{%- endif %}
{% if not options.dns64 -%}
"DNS64_DISABLE",
{%- endif %}
{% if not options.fallback -%}
"FALLBACK_DISABLE",
{%- endif %}
{%- endmacro %}

{% macro view_answer(answer) -%}
{%- if answer == 'allow' -%}
policy.TAGS_ASSIGN({})
{%- elif answer == 'refused' -%}
'policy.REFUSE'
{%- elif answer == 'noanswer' -%}
'policy.NO_ANSWER'
{%- endif -%}
{%- endmacro %}
